# PMML: Predictive Modelling Markup Language
#
# Part of the Rattle package for Data Mining
#
# Time-stamp: <2009-06-21 22:38:42 Graham Williams>
#
# Copyright (c) 2009 Togaware Pty Ltd
#
# This files is part of the Rattle suite for Data Mining in R.
#
# Rattle is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Rattle is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rattle. If not, see <http://www.gnu.org/licenses/>.

#######################################################################
# MAIN PMML FUNCTION

pmml <- function(model,
                 model.name="Rattle_Model",
                 app.name="Rattle/PMML",
                 description=NULL,
                 copyright=NULL,
                 transforms=NULL,
                 ...)
  UseMethod("pmml")

########################################################################
# UTILITY FUNCTIONS

generateCopyright <- function()
{
  return(paste("Copyright (c)", format(Sys.time(), "%Y"), Sys.info()["user"]))
}

sdecimal2binary <- function(x)
{
  return(rev(sdecimal2binary.smallEndian(x)))
}

sdecimal2binary.smallEndian <- function(x)
{
  if (x==0) return(0)
  if (x<0) stop("Sorry, the input must be positive")
  dec <- x

  n <- floor(log(x)/log(2))
  bin <- c(1)
  dec <- dec - 2 ^ n

  while(n > 0)
  {
    if (dec >= 2 ^ (n-1)) {bin <- c(bin,1); dec <- dec - 2 ^ (n-1)}
    else bin <- c(bin,0)
    n <- n - 1
  }
  return(bin)
}

# Function pmmlRootNode

pmmlRootNode <- function(version)
{
  # 080621 Don't use a default since we should specify in the call
  # which version the code is targetting - for documentation purposes.

  ##  PMML.VERSION <- "3.1" 080621 Did I user this string anywhere for any
  ##  external purpose. Maybe not - so it should be removed.
  
  # 080615 - The xmlns= namespace addition causes a warning on calling
  # getNodeSet unless we specify the namespace there:
  # 	> doc <- xmlTreeParse("test.xml", useInternalNodes=TRUE)
  #     > els <- getNodeSet(doc, "/PMML/DataDictionary/DataField")
  #     Warning message:
  #     using http://www.dmg.org/PMML-3_1 as prefix for default
  #     namespace http://www.dmg.org/PMML-3_1 
  #     > els <- getNodeSet(doc, "//p:DataField", c(p="http://www.dmg.org/PMML-3_1"))
  # We are supposed to include a default namespace, so do so. 

  if (version == "3.1")
    node <- xmlNode("PMML",
                    attrs=c(version="3.1",
                      xmlns="http://www.dmg.org/PMML-3_1",
                      "xmlns:xsi"="http://www.w3.org/2001/XMLSchema-instance"))
  else
    node <- xmlNode("PMML",
                    attrs=c(version="3.2",
                      xmlns="http://www.dmg.org/PMML-3_2",
                      "xmlns:xsi"="http://www.w3.org/2001/XMLSchema-instance", 
                      "xsi:schemaLocation"=paste("http://www.dmg.org/PMML-3_2",
                        "http://www.dmg.org/v3-2/pmml-3-2.xsd")))
  return(node)
}

pmmlHeader <- function(description, copyright, app.name)
{
  # Header
  
  VERSION <- "1.2.15" # Update documentation
  # "1.2.14" # Support mult transforms for rpart
  # "1.2.13" # Change strcutre used to record transforms.
  # "1.2.12" # Fix pmml.lm handling of singularities -> inactive
  # "1.2.11" # Fix categroics with one singularity in lm were marked inactive.
  # "1.2.10" # Fix typo in pmml.lm
  # "1.2.9" # Further fix a pmml.lm bug.
  # "1.2.8" # Fix a pmml.lm bug.
  # "1.2.7" # Export logistic classes
  # "1.2.6" # Support RMA transforms.
  # "1.2.5" # Include .TRANSFORM constants within pmml package.
  # "1.2.4" # Include collection of utility transform functions.
  # "1.2.3" # Bug fixes
  # "1.2.2" # Add test for transform support.
  # "1.2.1" # Streamline conditional handling of transforms.
  # "1.2.0" # Fix documentation and packaing and release to CRAN
  # "1.1.20" # Bug - fix rpart var names with transforms
  # "1.1.19" # Tidyup and update ClusterField
  # "1.1.18" # Include pmml.hclust in NAMESPACE
  # "1.1.17" # Export hclust as kmeans.
  # "1.1.16" # export pmml.multinom
  # "1.1.15" # Handle multinomial model.
  # "1.1.14" # Handle singularities in lm/glm better.
  # "1.1.13" # Support export of poisson(log)
  # "1.1.12" # Tree Array have quoted values. 0 for base in regression
  # "1.1.11" # Bug fix for pmml.lm - continuing to fix below problem
  # "1.1.10" # Bug fix for pmml.lm with categorical logistic target
  # "1.1.9" # Update rpart/nnet/ksvm from Zementis + many improvements
  # "1.1.8" # Increase number of digits extracted for rpart tests.
  # "1.1.7" # Add arules.
  # "1.1.6"
  # "1.1.5" # Add pmml.nnet.
  # "1.1.4" # Add pmml.ksvm. Fix extensions. 
  # "1.1.3" # Fixes for new version of randomSurvivalForest.
  # "1.1.2" Expose pmml.lm in NAMESPACE - woops.
  # "1.1.1" Add pmml.lm

  if (is.null(copyright)) copyright <- generateCopyright()
  header <- xmlNode("Header",
                    attrs=c(copyright=copyright, description=description))

  # Header -> Extension
						   
  header <- append.XMLNode(header,
                           xmlNode("Extension",
                                   attrs=c(name="timestamp",
                                     value=sprintf("%s", Sys.time()),
                                     extender="Rattle")))
  
  header <- append.XMLNode(header, xmlNode("Extension",
                                           attrs=c(name="description",
                                             value=sprintf("%s",
                                               Sys.info()["user"]),
                                             extender="Rattle")))
  

  # Header -> Application

  header <- append.XMLNode(header, xmlNode("Application",
                                           attrs=c(name=app.name,
                                             version=VERSION)))

  return(header)
}

pmmlDataDictionary <- function(field)
{
  # field$name is a vector of strings, and includes target
  # field$class is indexed by fields$names
  # field$levels is indexed by fields$names
  
  number.of.fields <- length(field$name)

  # DataDictionary

  data.dictionary <- xmlNode("DataDictionary",
                             attrs=c(numberOfFields=number.of.fields))
  data.fields <- list()
  for (i in 1:number.of.fields)
  {
    # Determine the operation type

    optype <- "UNKNOWN"
    datype <- "UNKNOWN"
    values <- NULL

    if (field$class[[field$name[i]]] == "numeric")
    {
      optype <- "continuous"
      datype <- "double"
    }
    else if (field$class[[field$name[i]]] == "factor")
    {
      optype <- "categorical"
      datype <- "string"
    }

    # DataDictionary -> DataField

     data.fields[[i]] <- xmlNode("DataField", attrs=c(name=field$name[i],
                                                optype=optype,
                                                dataType=datype))

    # DataDictionary -> DataField -> Value

    if (optype == "categorical")
      for (j in seq_along(field$levels[[field$name[i]]]))
        data.fields[[i]][[j]] <- xmlNode("Value",
                                         attrs=c(value=
                                           field$levels[[field$name[i]]][j]))
  }
  data.dictionary$children <- data.fields

  return(data.dictionary)
}

pmmlMiningSchema <- function(field, target=NULL, inactive=NULL)
{
  # 081103 Add inactive to list those variables that should be marked
  # as inactive in the model. This was added so that singularities can
  # be identified as inactive for a linear model. It could also be
  # used to capture ignored variables, if they were to ever be
  # included in the variable list.
  
  number.of.fields <- length(field$name)
  mining.fields <- list()

  for (i in 1:number.of.fields)
  {
    if (is.null(target))
      usage <- "active"
    else
      usage <- ifelse(field$name[i] == target, "predicted", "active")

    # 081103 Find out which variables should be marked as
    # inactive. Currently the inactive list is often supplied from
    # lm/glm as the variables which result in singularities in the
    # model. However, for categorics, this is the indicator variable,
    # like GenderMale. The test for %in% fails! So as a quick fix use
    # grep. This is not a solution (because the variable Test is a
    # substring of TestAll, etc). 

    # 090328 Revert to the exact test. We need to be cleverer in what
    # we pass through in the inactive vector. Whilst GenderMale might
    # be NA and thus is the only value included for this categoric,
    # for a categoric with more levels we need to no treat the others
    # as inactive so the whole categroic itself should not be
    # inactive. For now, the simple reversion works.
    
    if (field$name[i] %in% inactive) usage <- "inactive"
    # 090328 if (length(grep(field$name[i], inactive))) usage <- "inactive"
      
    mining.fields[[i]] <- xmlNode("MiningField",
                                  attrs=c(name=field$name[i],
                                    usageType=usage))
  }
  mining.schema <- xmlNode("MiningSchema")
  mining.schema$children <- mining.fields
  return(mining.schema)
}

########################################################################
# TRANSFORMS FRAMEWORK

# GLOBAL CONSTANTS

# 090517 List here all transforms that we know about from the rattle
# package. In using the pmml package with rattle these are used to
# identify tranformations. Information about how the transformation is
# performed (i.e., the transformation parameters) is separately
# recorded within rattle (crs$transforms) and must be used separately
# to generate the appropriate PMML code. Eventually I expect to see
# the handling of transform variables incorporated into the pmml
# package.

.TRANSFORMS.NORM.CONTINUOUS <- c("RRC", "R01", "RMD", "RMA")
.TRANSFORMS.IMPUTE <- c("IZR", "IMN", "IMD", "IMO", "ICN")
.TRANSFORMS.APPLY <- c("RLG")
.TRANSFORMS.BIN <- c("BQ", "BK", "BE")
.TRANSFORMS.INDICATOR <- c("TIN")

.TRANSFORMS.OTHER.NUM <- c("RRK", "TFC")
.TRANSFORMS.OTHER.CAT <- c("TJN", "TNM")
.TRANSFORMS.OTHER <- c(.TRANSFORMS.OTHER.NUM, .TRANSFORMS.OTHER.CAT)

.TRANSFORMS <- c(.TRANSFORMS.NORM.CONTINUOUS,
                 .TRANSFORMS.APPLY,
                 .TRANSFORMS.IMPUTE,
                 .TRANSFORMS.BIN,
                 .TRANSFORMS.INDICATOR,
                 .TRANSFORMS.OTHER)

supportTransformExport <- function(transforms=NULL)
{
  # 090208 Returns TRUE/FALSE. The default is that we don't have
  # Rattle code to generate the transforms in the pmml - i.e.,
  # pmml.transforms is not defined. A package which does will define a
  # pmml.transforms.
  
  return(exists("pmml.transforms") &&
         # length(getAnywhere("pmml.transforms")$objs) > 0 &&
         ! is.null(transforms))
}

unifyTransforms <- function(field, transforms)
{
  # 090102 Unify the list of variables in FIELD based on the known
  # TRANSFORMS, so that the variables in the TRANSFORMS are removed
  # from FIELD and replaced with the original variables from which the
  # TRANFROMS are derived. 090617 The final list of fields then are those
  # variables which need to be supplied to run the model.
  #
  # The variable TRANSFORMS is a list of transform structures (lists),
  # and each list has a type and original name.

  # 090111 We make sure that the field class is reset appropriately,
  # in special cases. So a binning transform, which is now a
  # categoric, was originally a numeric! Change it back to numeric.

  for (i in seq_along(transforms))
  {
    var <- transforms[i][[1]]
    type <- var$type
    vname <- names(transforms)[i] # The transformed variable name

    # 090102 The vname should be in the list of variables in field,
    # where we replace it with the original name. 090617 If it is not
    # in the list of variables then it should be an "intermediate"
    # transform, which is used later in another transform. We need to
    # handle this.

    # 090617 if (vname %in% field$name)
    # 090617 {
    index <- which(vname == field$name)
    if (length(index) > 0)
    {
          
      # 081229 I should probably be testing if index is integer(0) -
      # 090617 Now doing so.

      # 090607 REMOVE oname <- sub("^([^_]*)_", "", field$name[index])

      # 090111 For the binning operations, be sure to change the class
      # from categoric back to numeric.

      if (type %in% .TRANSFORMS.BIN)
        field$class[index] <- "numeric"

      # 090102 If the original variable (the untransformed variable)
      # for this transformed variable vname is already in the input
      # variables or is itself a transformed variable, then simply
      # remove the entry naming the transformed variable vname,
      # otherwise replace the entry naming the transformed var with
      # the original variable name.
      
      if (var$orig %in% union(field$name, names(transforms)))
      {
        field$name <- field$name[-index]
        field$class <- field$class[-index]
      }
      else
        field$name[index] <- var$orig
    }
    else if (length(which(var$orig == field$name)) == 0) # Orig not in field
    {
      field$name <- c(field$name, var$orig)
      field$class <- c(field$class, origVarType(var$type))
    }
  }

  # 090102 Reset the field$class names to correspond to the new
  # variables.

  names(field$class) <- field$name
  
  return(field)
}

pmmlCanExport <- function(vname)
{
  # 090517 In general this support function will return TRUE if the
  # variable name supplied is either not a transform (hence it is
  # exportable in general, since it is a supplied variable) or if it
  # is a transform, it is a supported transformation, if there is such
  # a list available. This is intended for the implementation of
  # transforms into PMML. An extermal package will provide a
  # pmml.transforms function and will override the simple default here
  # to test whether the transform is one that it can handle.
  return(! isTransformVar(vname))
}

isTransformVar <- function(vname)
{
# 090607
#  print(vname)
#  print(transformType(vname))
#  print(.TRANSFORMS)
#  cat(sprintf("Option 1: %s\n", transformType(vname) %in% .TRANSFORMS))
#  cat(sprintf("Option 2: %s\n", vname %in% names(crs$transforms)))
#  cat("Going with option 1\n")
  return(transformType(vname) %in% .TRANSFORMS)
# 090612 Revert to the above since crs$transforms is not available in
# the pmml package! This is used in pmmltocibi and ends up wrapping
# transforms with FloatVal in the wrong place.
#
#  print(vname %in% names(crs$transforms))
#  return(vname %in% names(crs$transforms))

  # 090613 It seems that returning FALSE always is okay! Thus need to
  # remove this throughout the code.

#  return(FALSE)
}

origVarType <- function(ty)
{
  if (ty %in% c(.TRANSFORMS.NORM.CONTINUOUS,
                .TRANSFORMS.IMPUTE, # 090617 But IZR and ICN are ambiguous.
                .TRANSFORMS.APPLY,
                .TRANSFORMS.BIN,
                .TRANSFORMS.OTHER.NUM))
    return("numeric")
  else
    return("factor")
}


transformType <- function(tr)
{
  # 090607 Return the code of the type of transform representated by
  # the variable TR. The supplied parameter TR must be a transform
  # data structure (list) and this function simply returns the "type"
  # element of the list. 090609 However, pmmltocibi currently calls
  # this function often with a strin g variable, so let's keep that
  # working for now until migrated.
  
  # 090517 Should we be returning NULL if there is no type recorded?
  # Or perhaps an empty string so that %in% will still work. 09607
  # Note that there should always be a type.

  if (class(tr) == "character") # 090709 Remove this sometime - needed
                                # in pmmltocibi for now.
  {
    code <- sub("^([^_]*)_.*$", "\\1", tr)
    if (substr(code, 1, 2) %in% .TRANSFORMS.BIN)
      code <- substr(code, 1, 2)
    return(code)
  }
  else
    return(tr[[1]]$type)
}

transformToDerived <- function(var)
{
  # 090607 REMOVE This function should not be needed any more with the
  # new transforms data structure (list). Remove it eventually, but
  # for now simply return the supplied var.

  return(var)
  
  if (transformType(var) %in% .TRANSFORMS.APPLY)
    # No args
    return(var) 
  else if (transformType(var) %in% .TRANSFORMS.IMPUTE)
    # One arg
    return(sub("^([^_]*_)(.*)(_[^_]*)$", "\\1\\2", var))
  else if (transformType(var) %in% .TRANSFORMS.NORM.CONTINUOUS)
    # two args
    return(sub("^([^_]*_)(.*)(_[^_]*){2}$", "\\1\\2", var))
  else if (transformType(var) %in% .TRANSFORMS.BIN)
  {
    nparms <- as.integer(sub("^..([^_]*)_.*$", "\\1", var)) + 1
    return(sub(sprintf("^([^_]*_)(.*)(_[^_]*){%s}$", nparms), "\\1\\2", var))
  }
  else if (transformType(var) %in% .TRANSFORMS.INDICATOR)
    # 090603 Split the values out and return list of transform var
    # names
    return(paste(transformToBasename(var), strsplit(var, "_")[[1]][-c(1,2)], sep="_"))
  else
    return(NULL)
}

transformToBasename <- function(var)
{
  # 090607 REMOVE No longer used. Simply return var. Remove this eventually.

  return(var)
  
  # 090603 This is awaiting turning crs$transforms into a structure to
  # handle this properly, then we won't need this kludge. For now we
  # assume no variable has _ in its name, but limit this assumption
  # to indicator variables for now.

  # 090601 For TIN_Marital_Married_..., return TIN_Marital.

  if (substr(var, 1, 4) == "TIN_")
    # 090603 TIN now lists all levels in one go.
    return(sub("^([^_]+_[^_]+)_.*$", "\\1", var))
    # 090601 Bit of an assumption that the levels do not have
    # underscores in them, but assume so for now. Strip off the level
    # from the transformed variable's name and return that.
    ## return(sub("^(.*)_[^_]+$", "\\1", var))
  else
    return(var)
}

transformToOriginal <- function(var)
{
  return(sub("^[^_]*_", "", transformToDerived(var)))
}

transformToParms <- function(var)
{
  if (transformType(var) %in% .TRANSFORMS.BIN)
  {
    nparms <- as.integer(sub("^..([^_]*)_.*$", "\\1", var)) + 1
    return(unlist(strsplit(
           sub("^ ", "",
           gsub("_", " ",
           sub(sprintf("^([^_]*_)(.*)((_[^_]*){%s})$",
                       nparms), "\\3", var))), " ")))
  }
}

